# pyOCD debugger
# Copyright (c) 2016-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core.memory_interface import MemoryInterface
from ..coresight.component import CoreSightCoreComponent
from ..coresight.cortex_m_core_registers import CortexMCoreRegisterInfo

class DebugContext(MemoryInterface):
    """@brief Viewport for inspecting the system being debugged.

    A debug context is used to access target registers and memory. It enables these accesses to be
    redirected to different locations. For instance, if you want to read registers from a call frame
    that is not the topmost, then a context would redirect those reads to locations on the stack.

    A context always has both a parent context and a specific core associated with it, neither of
    which can be changed after the context is created. The parent context is passed into the
    constructor. For the top-level debug context, the parent *is* the core. For other contexts that
    have a context as their parent, the core is set to the topmost parent's core.

    The DebugContext class itself is meant to be used as a base class. It's primary purpose is to
    provide the default implementation of methods to forward calls up to the parent and eventually
    to the core.
    """

    def __init__(self, parent):
        """@brief Debug context constructor.

        @param self
        @param parent The parent of this context. Can be either a core (CoreSightCoreComponent) or
            another DebugContext instance.
        """
        self._parent = parent

        if isinstance(self._parent, CoreSightCoreComponent):
            self._core = parent
        else:
            self._core = parent.core

    @property
    def parent(self):
        return self._parent

    @property
    def core(self):
        return self._core

    @property
    def session(self):
        return self.core.session

    def write_memory(self, addr, value, transfer_size=32):
        return self._parent.write_memory(addr, value, transfer_size)

    def read_memory(self, addr, transfer_size=32, now=True):
        return self._parent.read_memory(addr, transfer_size, now)

    def write_memory_block8(self, addr, value):
        return self._parent.write_memory_block8(addr, value)

    def write_memory_block32(self, addr, data):
        return self._parent.write_memory_block32(addr, data)

    def read_memory_block8(self, addr, size):
        return self._parent.read_memory_block8(addr, size)

    def read_memory_block32(self, addr, size):
        return self._parent.read_memory_block32(addr, size)

    def read_core_register(self, reg):
        """@brief Read one core register.

        @param self The debug context.
        @param reg Either the register's name in lowercase or an integer register index.
        @return The current value of the register. Most core registers return an integer value,
            while the floating point single and double precision register return a float value.

        @exception KeyError Invalid or unsupported register was requested.
        @exception @ref pyocd.core.exceptions.CoreRegisterAccessError "CoreRegisterAccessError" Failed to
            read the register.
        """
        reg_info = CortexMCoreRegisterInfo.get(reg)
        regValue = self.read_core_register_raw(reg_info.index)
        return reg_info.from_raw(regValue)

    def read_core_register_raw(self, reg):
        """@brief Read a core register without type conversion.

        @param self The debug context.
        @param reg Either the register's name in lowercase or an integer register index.
        @return The current integer value of the register. Even float register values are returned
            as integers (thus the "raw").

        @exception KeyError Invalid or unsupported register was requested.
        @exception @ref pyocd.core.exceptions.CoreRegisterAccessError "CoreRegisterAccessError" Failed to
            read the register.
        """
        vals = self.read_core_registers_raw([reg])
        return vals[0]

    def read_core_registers_raw(self, reg_list):
        """@brief Read one or more core registers.

        @param self The debug context.
        @param reg_list List of registers to read. Each element in the list can be either the
            register's name in lowercase or the integer register index.
        @return List of integer values of the registers requested to be read. The result list will
            be the same length as _reg_list_.

        @exception KeyError Invalid or unsupported register was requested.
        @exception @ref pyocd.core.exceptions.CoreRegisterAccessError "CoreRegisterAccessError" Failed to
            read one or more registers.
        """
        return self._parent.read_core_registers_raw(reg_list)

    def write_core_register(self, reg, data):
        """@brief Write a CPU register.

        @param self The debug context.
        @param reg The name of the register to write.
        @param data New value of the register. Float registers accept float values.

        @exception KeyError Invalid or unsupported register was requested.
        @exception @ref pyocd.core.exceptions.CoreRegisterAccessError "CoreRegisterAccessError" Failed to
            write the register.
        """
        reg_info = CortexMCoreRegisterInfo.get(reg)
        self.write_core_register_raw(reg_info.index, reg_info.to_raw(data))

    def write_core_register_raw(self, reg, data):
        """@brief Write a CPU register without type conversion.

        @param self The debug context.
        @param reg The name of the register to write.
        @param data New value of the register. Must be an integer, even for float registers.

        @exception KeyError Invalid or unsupported register was requested.
        @exception @ref pyocd.core.exceptions.CoreRegisterAccessError "CoreRegisterAccessError" Failed to
            write the register.
        """
        self.write_core_registers_raw([reg], [data])

    def write_core_registers_raw(self, reg_list, data_list):
        """@brief Write one or more core registers.

        @param self The debug context.
        @param reg_list List of registers to read. Each element in the list can be either the
            register's name in lowercase or the integer register index.
        @param data_list List of values for the registers in the corresponding positions of
            _reg_list_. All values must be integers, even for float registers.

        @exception KeyError Invalid or unsupported register was requested.
        @exception @ref pyocd.core.exceptions.CoreRegisterAccessError "CoreRegisterAccessError" Failed to
            write one or more registers.
        """
        self._parent.write_core_registers_raw(reg_list, data_list)

    def flush(self):
        self._core.flush()




# pyOCD debugger
# Copyright (c) 2021 mikisama
# Copyright (C) 2021 Ciro Cattuto <ciro.cattuto@gmail.com>
# Copyright (C) 2021 Simon D. Levy <simon.d.levy@gmail.com>
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List
import logging
from pyocd.core.helpers import ConnectHelper
from pyocd.core.memory_map import MemoryMap, MemoryRegion, MemoryType
from pyocd.core.soc_target import SoCTarget
from pyocd.subcommands.base import SubcommandBase
from pyocd.utility.cmdline import convert_session_options, int_base_0
from pyocd.utility.kbhit import KBHit
from ctypes import Structure, c_char, c_int32, c_uint32, sizeof


LOG = logging.getLogger(__name__)


class SEGGER_RTT_BUFFER_UP(Structure):
    """@brief `SEGGER RTT Ring Buffer` target to host."""

    _fields_ = [
        ("sName", c_uint32),
        ("pBuffer", c_uint32),
        ("SizeOfBuffer", c_uint32),
        ("WrOff", c_uint32),
        ("RdOff", c_uint32),
        ("Flags", c_uint32),
    ]


class SEGGER_RTT_BUFFER_DOWN(Structure):
    """@brief `SEGGER RTT Ring Buffer` host to target."""

    _fields_ = [
        ("sName", c_uint32),
        ("pBuffer", c_uint32),
        ("SizeOfBuffer", c_uint32),
        ("WrOff", c_uint32),
        ("RdOff", c_uint32),
        ("Flags", c_uint32),
    ]


class SEGGER_RTT_CB(Structure):
    """@brief `SEGGER RTT control block` structure. """

    _fields_ = [
        ("acID", c_char * 16),
        ("MaxNumUpBuffers", c_int32),
        ("MaxNumDownBuffers", c_int32),
        ("aUp", SEGGER_RTT_BUFFER_UP * 3),
        ("aDown", SEGGER_RTT_BUFFER_DOWN * 3),
    ]


class RTTSubcommand(SubcommandBase):
    """@brief `pyocd rtt` subcommand."""

    NAMES = ["rtt"]
    HELP = "SEGGER RTT Viewer."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""

        rtt_parser = argparse.ArgumentParser(cls.HELP, add_help=False)

        rtt_options = rtt_parser.add_argument_group("rtt options")
        rtt_options.add_argument("-a", "--address", type=int_base_0, default=None,
                                 help="Start address of RTT control block search range.")
        rtt_options.add_argument("-s", "--size", type=int_base_0, default=None,
                                 help="Size of RTT control block search range.")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, rtt_parser]

    def invoke(self) -> int:

        session = None
        kb = None

        try:
            session = ConnectHelper.session_with_chosen_probe(
                project_dir=self._args.project_dir,
                config_file=self._args.config,
                user_script=self._args.script,
                no_config=self._args.no_config,
                pack=self._args.pack,
                unique_id=self._args.unique_id,
                target_override=self._args.target_override,
                frequency=self._args.frequency,
                blocking=(not self._args.no_wait),
                connect_mode=self._args.connect_mode,
                options=convert_session_options(self._args.options))

            if session is None:
                LOG.error("No target device available")
                return 1

            with session:

                target: SoCTarget = session.board.target

                memory_map: MemoryMap = target.get_memory_map()
                ram_region: MemoryRegion = memory_map.get_default_region_of_type(MemoryType.RAM)

                if self._args.address is None or self._args.size is None:
                    rtt_range_start = ram_region.start
                    rtt_range_size = ram_region.length
                elif ram_region.start <= self._args.address and self._args.size <= ram_region.length:
                    rtt_range_start = self._args.address
                    rtt_range_size = self._args.size

                LOG.info(f"RTT control block search range [{rtt_range_start:#08x}, {rtt_range_size:#08x}]")

                data = target.read_memory_block8(rtt_range_start, rtt_range_size)
                pos = bytes(data).find(b"SEGGER RTT")

                if pos == -1:
                    LOG.error("No RTT control block available")
                    return 1

                rtt_cb_addr = rtt_range_start + pos

                rtt_cb = SEGGER_RTT_CB.from_buffer(bytearray(data[pos:]))
                up_addr = rtt_cb_addr + SEGGER_RTT_CB.aUp.offset
                down_addr = up_addr + sizeof(SEGGER_RTT_BUFFER_UP) * rtt_cb.MaxNumUpBuffers

                LOG.info(f"_SEGGER_RTT @ {rtt_cb_addr:#08x} with {rtt_cb.MaxNumUpBuffers} aUp and {rtt_cb.MaxNumDownBuffers} aDown")

                # some targets might need this here
                #target.reset_and_halt()

                target.resume()

                # set up terminal input
                kb = KBHit()

                # byte array to send via RTT
                cmd = bytes()

                while True:
                    # read data from up buffers (target -> host)    
                    data = target.read_memory_block8(up_addr, sizeof(SEGGER_RTT_BUFFER_UP))
                    up = SEGGER_RTT_BUFFER_UP.from_buffer(bytearray(data))

                    if up.WrOff > up.RdOff:
                        """
                        |oooooo|xxxxxxxxxxxx|oooooo|
                        0    rdOff        WrOff    SizeOfBuffer
                        """
                        data = target.read_memory_block8(up.pBuffer + up.RdOff, up.WrOff - up.RdOff)
                        target.write_memory(up_addr + SEGGER_RTT_BUFFER_UP.RdOff.offset, up.WrOff)
                        print(bytes(data).decode(), end="", flush=True)

                    elif up.WrOff < up.RdOff:
                        """
                        |xxxxxx|oooooooooooo|xxxxxx|
                        0    WrOff        RdOff    SizeOfBuffer
                        """
                        data = target.read_memory_block8(up.pBuffer + up.RdOff, up.SizeOfBuffer - up.RdOff)
                        data += target.read_memory_block8(up.pBuffer, up.WrOff)
                        target.write_memory(up_addr + SEGGER_RTT_BUFFER_UP.RdOff.offset, up.WrOff)
                        print(bytes(data).decode(), end="", flush=True)

                    else: # up buffer is empty

                        # try and fetch character
                        if not kb.kbhit():
                            continue
                        c = kb.getch()

                        if ord(c) == 8 or ord(c) == 127: # process backspace
                            print("\b \b", end="", flush=True)
                            cmd = cmd[:-1]
                            continue
                        elif ord(c) == 27: # process ESC
                            break
                        else:
                            print(c, end="", flush=True)
                            cmd += c.encode()

                        # keep accumulating until we see CR or LF
                        if not c in "\r\n":
                            continue

                        # SEND TO TARGET

                        data = target.read_memory_block8(down_addr, sizeof(SEGGER_RTT_BUFFER_DOWN))
                        down = SEGGER_RTT_BUFFER_DOWN.from_buffer(bytearray(data))

                        # compute free space in down buffer
                        if down.WrOff >= down.RdOff:
                            num_avail = down.SizeOfBuffer - (down.WrOff - down.RdOff)
                        else:
                            num_avail = down.RdOff - down.WrOff - 1

                        # wait until there's space for the entire string in the RTT down buffer
                        if (num_avail < len(cmd)):
                            continue

                        # write data to down buffer (host -> target), char by char
                        for i in range(len(cmd)):
                            target.write_memory_block8(down.pBuffer + down.WrOff, cmd[i:i+1])
                            down.WrOff += 1
                            if down.WrOff == down.SizeOfBuffer:
                                down.WrOff = 0;
                        target.write_memory(down_addr + SEGGER_RTT_BUFFER_DOWN.WrOff.offset, down.WrOff)

                        # clear it and start anew
                        cmd = bytes()

        except KeyboardInterrupt:
            pass

        finally:
            if session:
                session.close()
            if kb:
                kb.set_normal_term()

        return 0



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyocd.gdbserver.gdbserver import (
    escape,
    unescape,
)

# escaped chars: '#$}*'
# escaped by prefixing with '}' and xor'ing the char with 0x20
#
# '#' (0x23) -> '}\x03'
# '$' (0x24) -> '}\x04'
# '}' (0x7d) -> '}]'
# '*' (0x2a) -> '}\x0a'

class TestGdbServerEscaping:
    def test_escape_transparent(self):
        assert escape(b"hello") == b"hello"

    def test_escape_individual(self):
        assert escape(b"hello#foo") == b"hello}\x03foo"
        assert escape(b"hello$foo") == b"hello}\x04foo"
        assert escape(b"hello}foo") == b"hello}]foo"
        assert escape(b"hello*foo") == b"hello}\x0afoo"

    def test_escape_single(self):
        assert escape(b"#") == b"}\x03"
        assert escape(b"$") == b"}\x04"
        assert escape(b"}") == b"}]"
        assert escape(b"*") == b"}\x0a"

    def test_escape_combined(self):
        assert escape(b"#$}*") == b"}\x03}\x04}]}\x0a"
        assert escape(b'}}}') == b"}]}]}]"

    def test_unescape_transparent(self):
        assert unescape(b"bytes") == list(b"bytes")

    def test_unescape_individual(self):
        assert unescape(b"hello}\x03foo") == list(b"hello#foo")
        assert unescape(b"hello}\x04foo") == list(b"hello$foo")
        assert unescape(b"hello}]foo") == list(b"hello}foo")
        assert unescape(b"hello}\x0afoo") == list(b"hello*foo")

    def test_unescape_single(self):
        assert unescape(b"}\x03") == [b'#'[0]]
        assert unescape(b"}\x04") == [b'$'[0]]
        assert unescape(b"}]") == [b'}'[0]]
        assert unescape(b"}\x0a") == [b'*'[0]]

    def test_unescape_combined(self):
        assert unescape(b"}\x03}\x04}]}\x0a") == list(b"#$}*")
        assert unescape(b"}]}]}]") == list(b"}}}")



# pyOCD debugger
# Copyright (c) 2006-2013,2018 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import (Any, Optional, TYPE_CHECKING)

from ..core import exceptions
from ..target import TARGET
from ..target.pack import pack_target
from ..utility.graph import GraphNode

if TYPE_CHECKING:
    from ..core.session import Session

LOG = logging.getLogger(__name__)

class Board(GraphNode):
    """@brief Represents the board containing the target and associated components.

    The board is the root of the runtime object graph.
    """
    def __init__(self,
            session: "Session",
            target: Optional[str] = None,
            ) -> None:
        """@brief Constructor

        This method is responsible for selecting the SoCTarget subclass for the SoC, implementing the target
        type support. There are several possible sources for the target type name, with differing levels of
        priority.

        1. `target` parameter
        2. `target_override` session option
        3. `soft_target` parameter
        4. Last resort `cortex_m` target type. If this type is used, a warning is printed (unless the
            `warning.cortex_m_default` option is disabled).

        @param self
        @param session The session instance that owns us.
        @param target Target type name to use. If this parameter is set, it overrides all other sources of the
            target type.
        """
        super().__init__()

        # Use the session option if no target type was given to us.
        if target is None:
            target = session.options.get('target_override')

        # As a last resort, default the target to 'cortex_m'.
        if target is None:
            target = 'cortex_m'

            # Log a helpful warning when defaulting to the generic cortex_m target.
            if session.options.get('warning.cortex_m_default'):
                LOG.warning("Generic 'cortex_m' target type is selected by default; is this "
                            "intentional? You will be able to debug most devices, but not program "
                            " flash. To set the target type use the '--target' argument or "
                            "'target_override' option. Use 'pyocd list --targets' to see available "
                            "targets types.")

        assert target is not None

        # Convert dashes to underscores in the target type, and convert to lower case.
        target = target.replace('-', '_').lower()

        # Write the effective target type back to options if it's different.
        if target != session.options.get('target_override'):
            session.options['target_override'] = target

        self._session = session
        self._target_type = target
        self._test_binary = session.options.get('test_binary')
        self._delegate = None
        self._inited = False

        # Create targets from provided CMSIS pack.
        if session.options['pack'] is not None:
            pack_target.PackTargets.populate_targets_from_pack(session.options['pack'])

        # Create targets from the cmsis-pack-manager cache.
        if self._target_type not in TARGET:
            pack_target.ManagedPacks.populate_target(target)

        # Create Target instance.
        try:
            self.target = TARGET[self._target_type](session)
        except KeyError as exc:
            raise exceptions.TargetSupportError(
                f"Target type {self._target_type} not recognized. Use 'pyocd list --targets' to see currently "
                "available target types. "
                "See <https://github.com/pyocd/pyOCD/blob/master/docs/target_support.md> "
                "for how to install additional target support.") from exc

        # Tell the user what target type is selected.
        LOG.info("Target type is %s", self._target_type)

        self.add_child(self.target)

    def init(self) -> None:
        """@brief Initialize the board."""
        # If we don't have a delegate set yet, see if there is a session delegate.
        if (self.delegate is None) and (self.session.delegate is not None):
            self.delegate = self.session.delegate

        # Delegate pre-init hook.
        if (self.delegate is not None) and hasattr(self.delegate, 'will_connect'):
            self.delegate.will_connect(board=self)

        # Init the target.
        self.target.init()
        self._inited = True

        # Delegate post-init hook.
        if (self.delegate is not None) and hasattr(self.delegate, 'did_connect'):
            self.delegate.did_connect(board=self)

    def uninit(self) -> None:
        """@brief Uninitialize the board."""
        if self._inited:
            LOG.debug("uninit board %s", self)
            try:
                resume = self.session.options.get('resume_on_disconnect')
                self.target.disconnect(resume)
                self._inited = False
            except exceptions.Error as err:
                LOG.error("link exception during target disconnect: %s", err, exc_info=self._session.log_tracebacks)

    @property
    def session(self) -> "Session":
        """@brief The session that owns this board instance."""
        return self._session

    @property
    def delegate(self) -> Any:
        """@brief Delegate object that will be inherited by the SoCTarget."""
        return self._delegate

    @delegate.setter
    def delegate(self, the_delegate: Any) -> None:
        """@brief Set the delegate object that will be inherited by the SoCTarget."""
        self._delegate = the_delegate

    @property
    def unique_id(self) -> str:
        """@brief The probe's unique ID.

        Deprecated. Use the probe's `unique_id` property instead.
        """
        assert self.session.probe
        return self.session.probe.unique_id

    @property
    def target_type(self) -> str:
        """@brief Target type name."""
        return self._target_type

    @property
    def test_binary(self) -> Optional[str]:
        return self._test_binary

    @property
    def name(self) -> str:
        return "generic"

    @property
    def description(self) -> str:
        assert self.session.probe
        return "Generic board via " + self.session.probe.vendor_name + " " \
                + self.session.probe.product_name + " [" + self.target_type + "]"



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import (Any, Callable)

def locked(func: Callable) -> Callable:
    """@brief Decorator to automatically lock a method of a class.

    The class is required to have `lock()` and `unlock()` methods.
    """
    @wraps(func)
    def _locking(self, *args: Any, **kwargs: Any) -> Any:
        try:
            self.lock()
            return func(self, *args, **kwargs)
        finally:
            self.unlock()
    return _locking
